/*jslint indent:4 */
/*global define, module_, class_, include_, use_, extend_, static_, properties_, property_, globals_ */
/**
 * @projectDescription Event module
 *
 * @author Julien Polo<julien.polo@gmail.com>
 * @version 1.0.0
 */
define("stj/event", ["stj"], function () {
    "use strict";

    /**
     * @module event
     */
    module_(use_("event"), function () {

        /**
         * Shortcuts
         */
        var
            isUndefined = Object.isUndefined,
            slice = Array.prototype.slice.bindToCall(),
            forEach = Object.forEach,
            map = Object.map,
            size = Object.size,
            filter = use_('filter'),
            validateFunction = filter.validateFunction,
            validateNumber = filter.validateNumber,
            validateBoolean = filter.validateBoolean,
            validateObject = filter.validateObject,

            ValueError = use_('error').ValueError,
            eventNotDefined = function (eventName) {
                throw new ValueError({
                    "message" : "{eventName} is not defined",
                    "event" : eventName
                });
            };

        /**
         * @class event.Dispatcher
         */
        class_("Dispatcher", Object, function ($class) {
            extend_({
                LIMIT: 10
            });
            properties_({
                __size__: function __size__() {
                    return size(this.dispatcher);
                },

                /**
                 * @constructor
                 * @param {Object=} config
                 *  - limit : number of listeners allowed for one event
                 *  - strict: error will be thrown if event is not defined with defineEvent
                 *  - thisp: default object bound to listeners
                 */
                initialize: function initialize(config) {
                    this.dispatcher = {};
                    this.limit = $class.LIMIT;
                    this.strict = false;
                    this.thisp = this;
                    this.configure(config);
                },

                /**
                 * Destroy the dispatcher
                 */
                finalize: function finalize() {
                    this.dispatcher = null;
                },

                /**
                 * Configure the dispatcher
                 *
                 * @param {Object=}
                 */
                configure: function configure(config) {
                    this.defineEvent('addListener');
                    if (config) {
                        if (!isUndefined(config.limit)) {
                            this.limit = validateNumber(config.limit);
                        }
                        if (!isUndefined(config.strict)) {
                            this.strict = validateBoolean(config.strict);
                        }
                        if (!isUndefined(config.thisp)) {
                            this.thisp = config.thisp;
                        }
                    }
                },

                /**
                 * Add a new event type and return the array of listeners
                 *
                 * @param {String*}
                 * @return {Array}
                 */
                defineEvent: function defineEvent(type) {
                    var n = arguments.length - 1;
                    if (n >= 0) {
                        this._defineEvent(arguments[n]);
                        n -= 1;
                    }
                    return this;
                },

                /**
                 * Emit the event `type`
                 *
                 * @param {string} type
                 * @return {boolean} true if one ore more listener found
                 */
                emit: function emit(type) {
                    var args, useCall,
                        listeners = this.dispatcher[type], thisp,
                        listenersLength, i = 0,
                        handler, listener;

                    if (listeners) {
                        useCall = arguments.length <= 3;
                        args = useCall ? arguments : slice(arguments, 1);
                        thisp = this.thisp;

                        //listeners = slice(listeners);
                        listenersLength = size(listeners);

                        while (i < listenersLength) {
                            handler = listeners[i];
                            listener = handler[0];

                            //If once remove it
                            if (handler[1]) { //once test
                                listeners.splice(i, 1);
                            }

                            // fast cases
                            if (useCall) {
                                listener.call(thisp, args[1], args[2]);

                            // slower
                            } else {
                                listener.apply(thisp, args);
                            }

                            i += 1;
                        }
                        return true;
                    } else if (this.strict) {
                        eventNotDefined(type);
                    }
                    return false;
                },

                /**
                 * 
                 */
                getListeners: (function /*factory*/() {
                    var getFirstElement = function (array) {
                        return array[0];
                    };
                    return function getListeners(type) {
                        var listeners = this.dispatcher[type];
                        return listeners && map(listeners, getFirstElement);
                    };
                }()),

                /**
                 * Listen to event named `type`
                 *
                 * @param {string} type
                 * @param {Function} listener
                 * @param {boolean=} _once optional listening once option
                 * @return this
                 */
                addListener: function addListener(type, listener/*, _once*/) {
                    validateFunction(listener);
                    
                    var dispatcher = this.dispatcher,
                        //limit = this.limit,
                        listeners = dispatcher[type],
                        _once = !!arguments[2];

                    if (!listeners) {
                        if (this.strict) {
                            eventNotDefined(type);
                        }
                        listeners = this._defineEvent(type);
                    }


                    // To avoid recursion in the case that type == "newListeners"! Before
                    // adding it to the listeners, first emit "addListener".
                    this.emit('addListener', type, listener);

                    // Check for listener leak
                    /*if (!listeners.warned) {

                        if (limit && listeners.length > listenerMax) {
                            listeners.warned = true;
                            console.error('(node) warning: possible EventEmitter memory ' +
                                    'leak detected. %d listeners added. ' +
                                    'Use emitter.setMaxListeners() to increase limit.',
                                    listeners.length);
                            console.trace();
                        }
                    }*/

                    // If we've already got an array, just append.
                    listeners.push([listener, _once]);
                    return this;
                },

                /**
                 * Remove `listener` if listening to `type`
                 *
                 * @param {string} type
                 * @param {Function} listener
                 * @return {Function=} the listener if found
                 */
                removeListener: function removeListener(type, listener) {
                    validateFunction(listener);

                    // does not use listeners(), so no side effect of creating _events[type]
                    var listeners = this.dispatcher[type],
                        i;
                    if (listeners) {
                        i = size(listeners) - 1;
                        while (i >= 0) {
                            if (listeners[i][0] === listener) {
                                listeners.splice(i, 1);
                                return listener;
                            }
                            i -= 1;
                        }
                    } else if (this.strict) {
                        eventNotDefined(type);
                    }
                },

                /**
                 * Remove all listeners for events `type`
                 *
                 * @param {string} type
                 */
                removeAllListeners: function removeAllListeners(type) {
                    var dispatcher = this.dispatcher;
                    if (!type) {
                        forEach(dispatcher, function (_, key, self) {
                            self[key] = [];
                        });
                    } else if (dispatcher) {
                        dispatcher[type] = [];
                    }
                    return this;
                },

                _defineEvent: function _defineEvent(type) {
                    var dispatcher = this.dispatcher,
                        listeners = dispatcher[type];
                    if (!listeners) {
                        listeners = dispatcher[type] = [];
                    }
                    return listeners;
                }
            });

        });

        /**
         * @module event.Listenable
         */
        module_("Listenable", function () {

            properties_({
                emit: function emit(type) {
                    return this._getDispatcher().emit(type);
                },

                on: function on(type, listener) {
                    this._getDispatcher().addListener(type, listener, false);
                    return this;
                },

                once: function once(type, listener) {
                    this._getDispatcher().addListener(type, listener, true);
                    return this;
                },

                getListeners: function getListeners(type) {
                    return this._getDispatcher().getListeners(type);
                },

                addListener: function addListener(type, listener, _once) {
                    this._getDispatcher().addListener(type, listener, _once);
                    return this;
                },

                removeListener: function removeListener(type, listener) {
                    return this._getDispatcher().removeListener(type, listener);
                },

                removeAllListeners: function removeAllListeners(type) {
                    this._getDispatcher().removeAllListeners(type);
                    return this;
                },

                _getDispatcher: function _getDispatcher() {
                    var dispatcher = this.__dispatcher__,
                        DispatcherClass,
                        config;
                    if (!dispatcher) {
                        config = validateObject(this._getDispatcherConfig() || {});
                        config.thisp = config.thisp || this;
                        DispatcherClass = use_(config["class"] || 'event.Dispatcher');
                        dispatcher = this.__dispatcher__ = new DispatcherClass(config);

                        this._setupDispatcher(dispatcher);
                    }
                    return dispatcher;
                },

                // -"class"
                // - strict
                // - limit
                // - thisp
                _getDispatcherConfig: function _getDispatcherConfig() {
                    //override this method
                }

            });
        });
    });

    return use_('event');
});