/*jslint indent:4 */
/*global define, module_, class_, include_, use_, extend_, static_, properties_, property_, globals_ */
/**
 * @projectDescription inflection module
 *
 * @author Julien Polo<julien.polo@gmail.com>
 * @version 1.0.1
 */
define("stj/http", [ "stj", "std/Url", "std/Deferred" ], function () {
    "use strict";

    /**
     * @module http
     */
    module_("http", function ($module) {
        var isString = Object.isString,
            isUndefined = Object.isUndefined,
            merge = Object.update,

            filter = use_('filter'),
            validateObject = filter.validateObject,
            validateNumber = filter.validateNumber,
            
            Url = use_('Url'),

            getOptions = function (options) {
                if (isString(options)) {
                    options = {
                        url: options
                    };
                }
                return validateObject(options);
            };

        class_('Request', Object, function ($class) {

        });

        properties_({

            /**
             * Create a new request
             *
             * @param {String|Object} options
             * @param {Function=} callback
             * @return
             */
            request: function request(options, callback) {
                return $module._request(getOptions(options), callback);
            },

            /**
             * Configure the default options
             *
             * @param {Object} options
             *  - timeout
             *  - proxy
             *  - redirectLimit
             * @return this
             */
            configure: function configure(options) {
                var config = $module._config;
                if (options) {
                    if (!isUndefined(config.redirectLimit)) {
                        config.redirectLimit = validateNumber(options.redirectLimit);
                    }
                    if (!isUndefined(config.proxy)) {
                        config.proxy = Url.parse(options.proxy);
                    }
                }
                return $module;
            },

            /**
             * Shortcut for request with HEAD method
             *
             * @param {String|Object} options
             * @param {Function=} callback
             * @return
             */
            head: function head(options, callback) {
                options = getOptions(options);
                options.method = 'HEAD';
                if (options.body || options.json || options.multipart) {
                    throw new Error("HTTP HEAD requests MUST NOT include a request body.");
                }
                return $module._request(options, callback);
            },

            /**
             * Shortcut for request with GET method
             *
             * @param {String|Object} options
             * @param {Function=} callback
             * @return
             */
            get: function get(options, callback) {
                options = getOptions(options);
                options.method = 'GET';
                return $module._request(options, callback);
            },

            /**
             * Shortcut for request with POST method
             *
             * @param {String|Object} options
             * @param {Function=} callback
             * @return
             */
            post: function post(options, callback) {
                options = getOptions(options);
                options.method = 'POST';
                return $module._request(options, callback);
            },

            /**
             * Shortcut for request with PUT method
             *
             * @param {String|Object} options
             * @param {Function=} callback
             * @return
             */
            put: function put(options, callback) {
                options = getOptions(options);
                options.method = 'PUT';
                return $module._request(options, callback);
            },

            /**
             * Shortcut for request with DELETE method
             *
             * @param {String|Object} options
             * @param {Function=} callback
             * @return
             */
            del: function del(options, callback) {
                options = getOptions(options);
                options.method = 'DELETE';
                return $module._request(options, callback);
            },

            _config: {
                timeout: null,
                redirectMax: 10,
                proxy: null
            },
            _request: function _request(options, callback) {
                options = merge(getOptions(options), $module._config, false);
                var url = Url.parse(options.url),
                    proxy = options.proxy;

                if (proxy) {
                    options.port = proxy.port;
                    options.host = proxy.hostname;
                } else {
                    options.port = url.port;
                    options.host = url.hostname;
                }


                if (!url.port) {
                    if (url.protocol === 'http') {
                        url.port = 80;
                    } else if (url.protocol === 'https') {
                        url.port = 443;
                    }
                }

                if (proxy) {
                    options.path = url.toString();
                }

            }
        });
    });
});