/*jslint nodejs:true, indent:4 */
/*global define, module_, class_, include_, use_, extend_, static_, properties_, property_, globals_ */
/**
 * @projectDescription MIME type module
 *
 * @author Julien Polo<julien.polo@gmail.com>
 * @version 1.0.1
 */
define("stj/mime", [ "stj" ], function () {
    "use strict";

    /**
     * @module mime
     */
    module_("mime", function ($module) {
        var ValueError = use_('error.ValueError'),
            REGEXP_TYPE = /^[a-z0-9+.\/\-]+$/,
            REGEXP_EXTENSION = /^[a-z0-9]+$/;

        properties_({
            /**
             * Default value for unknown extensions
             */
            defaultType: "application/octet-stream",

            /**
             * Map of extension to mime type
             */
            _extensionToType: {},

            /**
             * Map of mime type to extension
             */
            _mimeToExtension : {},

            /**
             * Define mimetype -> extension mappings.  Each key is a mime-type that maps
             * to an array of extensions associated with the type.  The first extension is
             * used as the default extension for the type.
             *
             * e.g. mime.add({'audio/ogg': ['oga', 'ogg', 'spx']});
             *
             * @param {Object} map (Object) type definitions
             * @return this
             */
            add: function add(map) {
                var types = $module._extensionToType,
                    extensions = $module._mimeToExtension;


                Object.forEach(map, function (newExtensions, type) {
                    type = this._sanitizeMimetype(type);

                    var n = newExtensions.length - 1;
                    while (n >= 0) {
                        types[this._sanitizeExtension(newExtensions[n])] = type;
                        n -= 1;
                    }

                    //first extension is default extension
                    extensions[type] = newExtensions[0];
                }, $module);
                return $module;
            },

            /**
             * Lookup a mime type based on extension
             *
             * @param {string} extensionName
             * @param {string=} defaultValue
             * @return {string}
             */
            lookup: function lookup(extensionName, defaultValue) {
                var extensions = $module._extensionToType;
                return extensions[extensionName] || extensions[$module._sanitizeExtension(extensionName)] || defaultValue || $module.defaultType;
            },

            /**
             * Return file extension associated with a mime type
             *
             * @param {string} mimeType
             * @param {string}
             */
            extension: function extension(mimeType) {
                return this._mimeToExtension[$module._sanitizeMimeType(mimeType)];
            },

            /**
             * Lookup a charset based on mime type.
             */
            /*charsets: {
              lookup: function (mimeType, fallback) {
                // Assume text types are utf8.  Modify mime logic as needed.
                return /^text\//.test(mimeType) ? 'UTF-8' : fallback;
              }
            },*/

            _sanitizeExtension: function (extensionName) {
                extensionName = extensionName.toLowerCase();
                if (!REGEXP_EXTENSION.test(extensionName)) {
                    throw new ValueError(extensionName + " contains invalid characters");
                }
                return extensionName;
            },

            _sanitizeMimetype: function (mimeType) {
                mimeType = mimeType.toLowerCase();
                if (!REGEXP_TYPE.test(mimeType)) {
                    throw new ValueError(mimeType + " contains invalid characters");
                }
                return mimeType;
            }
        });

        /**
         * Load common mimetypes
         */
        $module.add({
            "application/atom+xml"          : ["atom"],
            "application/bson"              : ["bson"],
            "application/json"              : ["json"],
            "application/octet-stream"      : ["bin", "exe", "safariextz"],
            "application/pdf"               : ["pdf"],
            "application/postscript"        : ["ai", "eps", "ps"],
            "application/rtf"               : ["rtf"],
            "application/vnd.ms-fontobject" : ["eot"],
            "application/vnd.unity"         : ["unity3d"],
            "application/x-chrome-extension": ["crx"],
            "application/x-gtar"            : ["gtar"],
            "application/x-gzip"            : ["gz", "gzip"],
            "application/x-java-vm"         : ["class"],
            "application/x-java-archive"    : ["jar"],
            "application/x-javascript"      : ["javascript"],
            "application/x-opera-extension" : ["oex"],
            "application/x-shockwave-flash" : ["swf"],
            "application/x-tar"             : ["tar"],
            "application/x-xpinstall"       : ["xpi"],
            "application/zip"               : ["zip"],

            "audio/ogg"   : ["oga", "ogg", "spx"],
            "audio/x-wav" : ["wav"],
            "audio/x-midi": ["mid"],

            "font/otf"   : ["otf"],
            "font/ttf"   : ["ttf"],
            "font/x-woff": ["woff"],

            "image/bmp"    : ["bmp"],
            "image/gif"    : ["gif"],
            "image/jpeg"   : ["jpg", "jpeg", "jpe"],
            "image/png"    : ["png"],
            "image/svg+xml": ["svg", "svgz"],
            "image/targa"  : ["tga"],
            "image/tiff"   : ["tif", "tiff"],
            "image/webp"   : ["webp"],
            "image/x-icon" : ["ico"],
            "image/x-jng"  : ["jng"],

            "text/cache-manifest": ["appcache", "manifest"],
            "text/html"          : ["htm", "html"],
            "text/plain"         : ["txt", "log"],
            "text/richtext"      : ["rtf", "rtx"],
            "text/xml"           : ["rss"],
            "text/yaml"          : ["yaml"],

            "video/mpeg"     : ["mpg", "mpeg", "mpe"],
            "video/quicktime": ["qt", "mov"],
            "video/ogg"      : ["ogv"],
            "video/webm"     : ["webm"],
            "video/x-msvideo": ["avi"]
        });
    });

    return use_('mime');
});
